unit Map;

interface

uses
  SysUtils, IdGlobal, Token, MyRobot, Packages;

type
  RRobot = Record
    Id: Integer;                // Robot's ID
    LocX, LocY: Integer;        // Location
    isAlive: Boolean;           // Alive;
    packageCount: Integer;      // Package Count;
  end;

type
  MapObject = Record
    flag: Integer;
    robot: Integer;
    robotPacks: Integer;
    packages: Boolean;
    myPacks: TPackages;
    gatherPoint: Boolean;
    tempBlock: Boolean;
  end;

type
  RHomeBase = Record
    locX, locY: Integer;        // Location of the home base
    lastChecked: Integer;       // Last time we checked it (money wise)
    packageCount: Integer;      // Number of Packages at last count;
  end;

type
  TMap = Class (TObject)
  public
    constructor Create(Data: String);
    destructor Destroy(); override;
    procedure ParseLine(Data: String);
    procedure ParseGameLine(var Data: String);
    procedure ParsePackageLine(Data: String);
    procedure CreateMyPlayer(Data: String);
    procedure DisplayMap(cols, rows: Integer);

    procedure robotAI();

    function haveAllData(): Boolean;
    function getPath(pStart, pEnd: Point; rnd, temp: Boolean): String;
    function amIAlone(dis: Integer): Boolean;
    function canDestroy(var dir: char): Boolean;

    function aiCouldDie(): Boolean;

    function getDisplayLine1(): String;
    function getDisplayLine2(): String;
    function getDisplayLine3(): String;
    function getDisplayLine4(): String;
    function getDisplayLine5(): String;
    function getDisplayLine6(): String;
    function getDisplayLine7(): String;
    function getDisplayLine8(): String;
    function getDisplayLine9(): String;
    procedure padString(var line: String; len: Integer);
    function getMapChar(x,y: Integer; simple: Boolean): String;

    function BoolToStr(b: Boolean): String;
  private
    width, height: Integer;          // Width and Height of the map
    lineOn: Integer;
    amSetup: Boolean;
    canDisplay: Boolean;
    pkSet: String;

    // 2D array for our map Objects
    mapData: array of array of MapObject;

    // My Robot
    myRobot: TMyRobot;

    // Other Robots
    robotList: array of RRobot;

    // Home Bases
    baseList: array of RHomeBase;

    // Make sure we don't bid 2 for too lone (pushing another bot)
    push2Count: Integer;

    // How far away we hide things
    hideDistance: Integer;

    procedure ParsePlayerMove(Data: String);

    // Finds Shore Lines (higher movement cost so we can avoid them if posible)
    procedure findShores();
    function isShore(x, y: Integer): Boolean;

    // AI Sub Functions
    procedure aiFindPackage();
    procedure aiDeliverPackage();
    procedure aiGatherPut();
    procedure aiOblivion();
    procedure aiBidLevel();
    function aiPushBlocked(): Boolean;
    function aiGetSafeDir(): String;
    function aiPushToOblivion(var Path: String): Boolean;

    // Robot List Commands
    procedure addRobot(id: Integer);
    function haveRobot(id: Integer): Boolean;
    procedure robotSetX(id, x: Integer);
    procedure robotSetY(id, y: Integer);
    procedure robotMove(id: Integer; dir: Char);
    procedure robotPick(id, pId: Integer);
    procedure robotDrop(id, pId: Integer);
    procedure robotAlive(id: Integer);
    procedure robotKillAll();
    function robotCount(): Integer;

    function findClosestPackage(pStart: Point; var path: String): Point;
    function findHomeBase(pStart: point; var path: String): Point;
    procedure findDeliveryDist(tempPacks: TPackages);

    procedure addHomeBase(x,y: Integer);
    procedure setHomeBase(x, y, cTime, pCount: Integer);
  published
    property Setup: Boolean read amSetup;
    property Display: Boolean read canDisplay;
    property me: TMyRobot read myRobot;
    property packDebug: String read pkSet;
    property mapWidth: Integer read width;
    property mapHeight: Integer read height;
end;


implementation

const
  FLAG_UNKNOWN = $FF;
  FLAG_PLAIN = $00;
  FLAG_HOME = $01;
  FLAG_SHORE = $02;
  FLAG_WATER = $10;
  FLAG_WALL = $11;

// Sets up our map space, Data is the first line and gives us
// the width and height of the map;
constructor TMap.Create(Data: String);
var
  i, j: Integer;
  myToken: TToken;
begin
  inherited Create();

  Randomize();
  pkSet := '';
  push2Count := 0;

  myToken := TToken.Create(Data, ' ', '+');
  if (myToken.countTokens = 2) then begin
    LineOn := 0;        // We haven't parsed any lines (map data) yet
    width := StrToInt(myToken.nextToken());
    height := StrToInt(myToken.nextToken());
    hideDistance := Max((((2*width) + (2*height)) shr 7) + 2, 4);
    setLength(mapData, height, width);
    WriteLn('Parsing Map Data....');
    amSetup := true;
    canDisplay := false;
    setLength(robotList, 0);
    // Decrease width and height by 1 so we don't have to do it in loops
    dec(height);
    dec(width);
    for i := 0 to height do begin
      for j := 0 to width do begin
        mapData[i][j].flag := FLAG_UNKNOWN;
        mapData[i][j].robot := -1;
        mapData[i][j].packages := false;
      end;
    end;
  end
  else begin
    WriteLn('Malformed Map Constructor, first line must be map width height');
    amSetup := false;
    canDisplay := false;
  end;
end;

// Destroys us and all the map data we have, must play nice with memory.
Destructor TMap.Destroy();
var i, j: Integer;
begin
  myRobot.Free();
  for i := 0 to height do begin
    for j := 0 to width do begin
      mapData[i][j].myPacks.Free();
    end;
  end;

  inherited Destroy();
end;

Procedure TMap.ParseLine(Data: String);
var
  i: Integer;
  c: Char;
begin
  // Make sure we're setup before blindly parsing the data
  if (amSetup) then begin
    if (length(Data) = width + 1) then begin
      for i := 0 to width do begin
        c := Data[i+1];
        mapData[lineOn][i].robot := -1;
        mapData[lineOn][i].robotPacks := 0;
        mapData[lineOn][i].packages := false;
        mapData[lineOn][i].myPacks := nil;
        mapData[lineOn][i].gatherPoint := false;
        mapData[lineOn][i].tempBlock := false;
        case (c) of
          '.': MapData[lineOn][i].flag := FLAG_PLAIN;
          '@': begin
                 MapData[lineOn][i].flag := FLAG_HOME;
                 MapData[lineOn][i].packages := true;
                 addHomeBase(i, lineOn);
               end;
          '~': MapData[lineOn][i].flag := FLAG_WATER;
          '#': MapData[lineOn][i].flag := FLAG_WALL;
        end;
      end;
      inc(lineOn);
      if (lineOn = height + 1) then
        findShores();
    end
    else begin
      WriteLn('Malformed map data on line ' + IntToStr(lineOn) + '!!!!');
    end;
  end;
end;

// This just splits our data into individual player moves by splitting
// by the substring #.
procedure TMap.ParseGameLine(var Data: String);
var myToken: TToken;
begin
  // Remove starting #
  Data := Copy(Data, 2, length(Data));

  // Create the tokenizer and parse every player's move
  myToken := TToken.Create(Data, '#', '+');
  robotKillAll();
  while (myToken.hasMoreTokens) do begin
    self.ParsePlayerMove(myToken.nextToken());
  end;
  myToken.free();
end;

procedure TMap.ParsePackageLine(Data: String);
var
  tempPacks: TPackages;
  path: String;
  dest: Point;
begin
  if (length(Data) > 0) then begin
    tempPacks := mapData[myRobot.location.y][myRobot.location.x].myPacks;
    if (tempPacks = nil) then
      tempPacks := TPackages.Create(Data)
    else
      tempPacks.buildPackageList(Data);
    mapData[myRobot.location.y][myRobot.location.x].packages := true;
    mapData[myRobot.location.y][myRobot.location.x].myPacks := tempPacks;

    // Now that we have the package list we need to find the distance
    // to the diliery points
    findDeliveryDist(tempPacks);

    setHomeBase(myRobot.location.x, myRobot.location.y, myRobot.cash, tempPacks.getCount());

    // Check to see if this data was expected, if not and we're look
    // for packages more then 5 spaces away, go here!
    if (myRobot.state = ROBOT_FINDPACKAGE) and
       (length(myRobot.path) > 5) then begin
      path := '';
      dest := findClosestPackage(myRobot.location, path);
      if (dest.x >= 0) and (dest.y >= 0) then begin
        myRobot.setDest(dest, path);
      end;
    end;
  end
  else begin
    mapData[myRobot.location.y][myRobot.location.y].packages := false;
    setHomeBase(myRobot.location.x, myRobot.location.y, myRobot.cash, 0);
  end;;
end;

procedure TMap.CreateMyPlayer(Data: String);
begin
  myRobot := TMyRobot.Create(Data);
  canDisplay := true;
end;

procedure TMap.DisplayMap(cols, rows: Integer);
var
  sx, sy: Integer;
  ex, ey: Integer;
  i, j: Integer;
  line: String;
begin
  if (amSetup) and (lineOn = height+1) then begin
    sx := myRobot.location.x - (cols div 2);
    sy := myRobot.location.y - (rows div 2);
    ex := sx + cols;
    ey := sy + rows;
    for i := sy to ey do begin
      line := '';
      for j := sx to ex do begin
        line := line + getMapChar(j, i, false);
      end;
      WriteLn(line);
    end;
  end;
end;

// This function does all the heavy lifting when it comes to
// performing AI functionality.
procedure TMap.robotAI();
begin
  myRobot.alone := amIAlone(2);
  myRobot.pushAlone := amIAlone(1);
  myRobot.veryAlone := amIAlone(12);
  myRobot.oponents := robotCount();

  if (myRobot.state = ROBOT_DELIVER) and (myRobot.haveGatherDest) then
    myRobot.setStopGathering();

  if (myRobot.state = ROBOT_DELIVER) then begin
    aiDeliverPackage();
  end
  else if (myRobot.state = ROBOT_FINDPACKAGE) then begin
    aiFindPackage();
  end
  else if (myRobot.state = ROBOT_HIDE_GET) then begin
    aiFindPackage();
  end
  else if (myRobot.state = ROBOT_HIDE_PUT) then begin
    aiGatherPut();
  end;

  // Make sure package pickup was good, if not it wasn't kill package info
  if (myRobot.pickupFailed) then begin
    mapData[myRobot.location.y][myRobot.location.x].myPacks.Free();
    mapData[myRobot.location.y][myRobot.location.x].myPacks := nil;
    mapData[myRobot.location.y][myRobot.location.x].packages := false;
  end;

  // Set bidding, good for all states
  aiBidLevel();

  // If we can kill, that wins out over every other command
  // except for us getting killed.
  if (aiCouldDie()) then begin
    myRobot.bid := myRobot.getAvgDeliverCost() + (myRobot.cash shr 4) + 20;
    myRobot.move := 'Move';
    myRobot.moveData := aiGetSafeDir();
  end else begin
    aiOblivion();
  end
end;

// Returns true if we've read all our lines, false otherwise
function TMap.haveAllData(): Boolean;
begin
  Result := (lineOn = (height+1));
end;

// Trys to find the shortest path between the two points
//
// Works from the end point and building a movement cost
// array.  The movement cost at the start point is always
// the lowest value posible and a path will always be found.
//
// The only problem with this is that it can take quite some
// time to run on larger maps and it can eat a lot of stack space.
// To prevent a runaway loop, we'll limit our steps to the
// map width + map height.
function TMap.getPath(pStart, pEnd: Point; rnd, temp: Boolean): String;
var
  maxSteps, steps: Integer;     // Maximum number of steps to calculate
  markMap: array of array of Integer;   // Our Mark Map
  i, j: Integer;        // For counting through all map data
  found: Boolean;       // Makes sure we've found a next step
  sx, sy: Integer;      // Start x, start Y looping;
  ex, ey: Integer;      // end x, end y looping
  atX, atY: Integer;    // At X and At Y, for walking the path
  lowvalue: Integer;    // Lowest Distance
  noGo: Int64;          // Highest posible path length + 1
  charToAdd: String;
begin
  if (mapData[pEnd.y][pEnd.x].flag >= FLAG_WATER) or
     (mapData[pEnd.y][pEnd.x].robot > 0) then begin
    Result := '!!! No Path !!!';
  end
  else begin
    Result := '';
    // maxSteps := (width+1) + (height+1) ;
    noGo := (width+1) * (height+1) + 1;
    maxSteps := noGo shl 3;
    setLength(markMap, height+1, width+1);
    { Fill up the mark Map }
    for i := 0 to height do begin
      for j := 0 to width do begin
        { Water, walls or robots cannot be passed }
        if (j = myRobot.location.x) and
           (i = myRobot.location.y) then
          markMap[i][j] := -1
        else if (mapData[i][j].flag > $0F) or
           (mapData[i][j].robot <> -1) then
          markMap[i][j] := noGo
        else if (temp) and (mapData[i][j].tempBlock) then
          markMap[i][j] := noGo
        else if (mapData[i][j].flag = FLAG_SHORE) then
          markMap[i][j] := -4
        else
          markMap[i][j] := -1;
      end;
    end;

    markMap[pEnd.y][pEnd.x] := 0;
    sx := pEnd.x;         // Set our start x to end point
    sy := pEnd.y;         // Set our start y to end point
    ex := pEnd.x;         // Set our end x to end point
    ey := pEnd.y;         // Set out end y to end point
    steps := 0;           // Current number of steps
    found := true;

    // Find out path, keep adding
    while (found) and (steps < maxSteps) do begin
      found := false;
      for i := sy to ey do begin
        for j := sx to ex do begin
          // Only scan if our current location is passable and has a set value
          if (markMap[i][j] <> noGo) and (markMap[i][j] > -1) then begin
            // Scan if Moving North is allowed, if so set mark map
            if (markMap[Max(i-1,0)][j] < 0) then begin
              found := true;
              markMap[Max(i-1,0)][j] := markMap[i][j] - markMap[Max(i-1,0)][j];
            end;
            // Scan if Moving South is allowed, if so set mark map
            if (markMap[Min(i+1,height)][j] < 0) then begin
              found := true;
              markMap[Min(i+1,height)][j] := markMap[i][j] - markMap[Min(i+1,height)][j];
            end;
            // Scan if Moving West is allowed, if so set mark map
            if (markMap[i][Max(j-1,0)] < 0) then begin
              found := true;
              markMap[i][Max(j-1,0)] := markMap[i][j] - markMap[i][Max(j-1,0)];
            end;
            // Scan if Moving East is allowed, if so set mark map
            if (markMap[i][Min(j+1,width)] < 0) then begin
              found := true;
              markMap[i][Min(j+1,width)] := markMap[i][j] - markMap[i][Min(j+1,width)];
            end;
          end;
        end;
      end;
      sx := Max(sx-1, 0);         // Set Our start x to start x - 1 or map edge
      sy := Max(sy-1, 0);         // Set out start y to start y - 1 or map edge
      ex := Min(ex+1, width);     // Set out end x to end y + 1 or map edge
      ey := Min(ey+1, height);    // Set out end y to end y + 1 or map edge
      inc(steps);
    end;

    // Now we have check to see if a path was found and build it
    atX := pStart.x;
    atY := pStart.y;
    lowValue := markMap[atY][atX];

    if (markMap[atY][atX] <> -1) and (markMap[atY][atX] <> noGo) then begin
      while (lowValue > 0) do begin
        charToAdd := '';
        // Randomize if N/S or E/W is checked first
        if (not rnd) or (Random(100) > 50) then begin
          // Check to see if moving North gets us closer to our end point
          if (atY-1 >= 0) and
             (markMap[atY-1][atX] < lowValue) then begin
            charToAdd := STMOVE_NORTH;
            lowValue := markMap[atY-1][atX];
          end;
          // Check to see if moving South gets us closer to our end point
          if (atY+1 <= height) and
             (markMap[atY+1][atX] < lowValue) then begin
            charToAdd := STMOVE_SOUTH;
            lowValue := markMap[atY+1][atX];
          end;
          // Check to see if moving West gets us closer to out end point
          if (atX-1 >= 0) and
             (markMap[atY][atX-1] < lowValue) then begin
            charToAdd := STMOVE_WEST;
            lowValue := markMap[atY][atX-1];
          end;
          // Check to see if moving East gets us closer to out end point
          if (atX+1 <= width) and
             (markMap[atY][atX+1] < lowValue) then begin
            charToAdd := STMOVE_EAST;
            lowValue := markMap[atY][atX+1];
          end;
        end
        else begin
          // Check to see if moving West gets us closer to out end point
          if (atX-1 >= 0) and
             (markMap[atY][atX-1] < lowValue) then begin
            charToAdd := STMOVE_WEST;
            lowValue := markMap[atY][atX-1];
          end;
          // Check to see if moving East gets us closer to out end point
          if (atX+1 <= width) and
             (markMap[atY][atX+1] < lowValue) then begin
            charToAdd := STMOVE_EAST;
            lowValue := markMap[atY][atX+1];
          end;
          // Check to see if moving North gets us closer to our end point
          if (atY-1 >= 0) and
             (markMap[atY-1][atX] < lowValue) then begin
            charToAdd := STMOVE_NORTH;
            lowValue := markMap[atY-1][atX]
          end;
          // Check to see if moving South gets us closer to our end point
          if (atY+1 <= height) and
             (markMap[atY+1][atX] < lowValue) then begin
            charToAdd := STMOVE_SOUTH;
            lowValue := markMap[atY+1][atX]
          end;
        end;

        // Walk
        Result := Result + charToAdd;
        if (charToAdd = STMOVE_NORTH) then
          dec(atY)
        else if (charToAdd = STMOVE_SOUTH) then
          inc(atY)
        else if (charToAdd = STMOVE_WEST) then
          dec(atX)
        else if (charToAdd = STMOVE_EAST) then
          inc(atX);

        // Update our low value;
        lowValue := markMap[atY][atX];
      end;
    end
    else begin
      Result := '!!! No Path !!!';
    end;
  end;
end;

// Walk the given path and returns the end point
function getWalk(ptStart: Point; path: String): Point;
var i: Integer;
begin
  Result := ptStart;
  for i := 1 to length(path) do begin
    case (path[i]) of
      STMOVE_NORTH:       dec(Result.y);
      STMOVE_SOUTH:       inc(Result.y);
      STMOVE_WEST:        dec(Result.x);
      STMOVE_EAST:        inc(Result.x);
    end;
  end;
end;

// This function determins if player's robot is alone (no one)
// within range of pushing no mater how they move.  If we are
// alone then our bid really doesn't matter 1 will be just fine.
function TMap.amIAlone(dis: Integer): Boolean;
var
  sx,sy: Integer;       // Starting x, y
  ex,ey: Integer;       // Ending x,y
  i, j: Integer;
  checkDis: Integer;    // Check Out Distance
  loc: Point;           // Player's Location
begin
  Result := true;
  if (amSetup and myRobot.ready) then begin
    loc := myRobot.location;
    sx := Max(loc.x-dis,0);
    sy := Max(loc.y-dis,0);
    ex := Min(loc.x+dis,width);
    ey := Min(loc.y+dis,height);
    for i := sy to ey do begin
      for j := sx to ex do begin
        // Distance check, no need to check if the distance is over the given
        // This oddly enough takes longer to compute but is more accurate
        checkDis := Abs(j-loc.x) + Abs(i-loc.y);
        // Check to see if there is another robot in range, if so, find it!
        if (mapData[i][j].robot > 0) and (checkDis <= dis) and
           (mapData[i][j].robot <> myRobot.identifier) then
          Result := false;
      end;
    end;
  end;
end;

// This function, much like amAlone, checks to see if we have
// the chance to destroy a player (push them into water).  If we
// can, the dir variable it set to the direction we need to move.
function TMap.canDestroy(var dir: Char): Boolean;
var
  loc: Point;           // Player's Location
begin
  Result := false;
  if (amSetup and myRobot.ready) then begin
    // Check a push to the north
    if (loc.y-2 >=0) and
       (mapData[loc.y-1][loc.x].robot > 0) and
       (mapData[loc.y-2][loc.x].flag = FLAG_WATER) then begin
      Result := true;
      dir := STMOVE_NORTH;
    end
    // Check a push to the south
    else if (loc.y+2 <= height) and
            (mapData[loc.y+1][loc.x].robot > 0) and
            (mapData[loc.y+2][loc.x].flag = FLAG_WATER) then begin
      Result := true;
      dir := STMOVE_SOUTH;
    end
    // Check a push to the west
    else if (loc.x-2 >= 0) and
            (mapData[loc.y][loc.x-1].robot > 0) and
            (mapData[loc.y][loc.x-2].flag = FLAG_WATER) then begin
      Result := true;
      dir := STMOVE_WEST;
    end
    // Check a push to the east
    else if (loc.x+2 <= width) and
            (mapData[loc.y][loc.x+1].robot > 0) and
            (mapData[loc.y][loc.x-2].flag = FLAG_WATER) then begin
      Result := true;
      dir := STMOVE_EAST;
    end;
  end;
end;


procedure TMap.ParsePlayerMove(Data: String);
var
  myToken: TToken;
  myBotId: Integer;
  move: Char;
  iData: Integer;
begin
  myToken := TToken.Create(Data, ' ', '+');
  myBotId := StrToInt(myToken.nextToken());
  if (myBotId <> myRobot.identifier) then
    robotAlive(myBotId);
  while (myToken.hasMoreTokens) do begin
    move := myToken.nextToken()[1];
    if (myBotId = myRobot.identifier) then begin
      myRobot.lastAction := Data;
      try
        case (move) of
          'X':  myRobot.setLocationX(StrToInt(myToken.nextToken()));
          'Y':  myRobot.setLocationY(StrToInt(myToken.nextToken()));
          'S':  myRobot.doMove('S');
          'N':  myRobot.doMove('N');
          'E':  myRobot.doMove('E');
          'W':  myRobot.doMove('W');
          'P':  begin
                  iData := StrToInt(myToken.nextToken());
                  robotPick(-1, iData);
                  myRobot.doPick(iData);
                end;
          'D':  myRobot.doDrop(StrToInt(myToken.nextToken()));
        end;
      finally
      end;
    end else begin
      if (not haveRobot(myBotID)) then
        addRobot(myBotID);

      try
        case (move) of
          'X':  robotSetX(myBotID, StrToInt(myToken.nextToken()));
          'Y':  robotSetY(myBotID, StrToInt(myToken.nextToken()));
          'S':  robotMove(myBotID, 'S');
          'N':  robotMove(myBotID, 'N');
          'E':  robotMove(myBotID, 'E');
          'W':  robotMove(myBotID, 'W');
          'P':  robotPick(myBotID, StrToInt(myToken.nextToken()));
          'D':  robotDrop(myBotID, StrToInt(myToken.nextToken()));
        end;
      finally
      end;
    end;
  end;
  myToken.Free();
end;

function TMap.findClosestPackage(pStart: point; var path: String): Point;
var
  i,j: Integer;         // Counters
  pEnd: Point;          // End Point (Where the packages are)
  lowDis: Integer;      // Lowest Distance found so far
  tempPath: String;
begin
  Result.x := -1; Result.y := -1;
  lowDis := -1;
  for i := 0 to height do begin
    for j := 0 to width do begin
      // If we have packages but can't see to pick up anything solve
      // for distances again.
      if (mapData[i][j].myPacks <> nil) and
         (mapData[i][j].myPacks.getPickupWeight(myRobot.capacity,
                (mapData[i][j].flag = FLAG_HOME)) = 0) and
         (mapData[i][j].myPacks.getCount() > 0) then
        findDeliveryDist(mapData[i][j].myPacks);
      if (mapData[i][j].packages) or
         ((mapData[i][j].myPacks <> nil) and
          (mapData[i][j].myPacks.getPickupWeight(myRobot.capacity,
                (mapData[i][j].flag = FLAG_HOME)) > 0)) then begin
        pEnd.x := j;
        pEnd.y := i;
        tempPath := getPath(pStart, pEnd, true, false);
        if (length(tempPath) > 0) and (tempPath[1] <> '!') then begin
          if (lowDis = -1) or (length(tempPath) < lowDis) then begin
            lowDis := length(tempPath);
            path := tempPath;
            Result := pEnd;
          end;
        end;
      end;
    end;
  end;
  if (lowDis = -1) then
    Result := findHomeBase(pStart, path);
end;

// If we can't find a place where we know packages are it's time to
// move to the home base it's been longest since we checked.
function TMap.findHomeBase(pStart: point; var path: String): Point;
var
  i: Integer;
  highTime: Integer;    // Highest Time found
  pEnd: Point;          // End Point (Where the packages are)
  found: Boolean;       // Found a new home base to scan
  tempPath: String;
begin
  highTime := -2;
  found := false;
  for i := low(baseList) to high(baseList) do begin
    if (highTime = -1) then begin { Do Nothing } end
    else if (highTime = -2) or
       (baseList[i].lastChecked = -1) or
       (baseList[i].lastChecked > highTime) then begin
      pEnd.x := baseList[i].locX;
      pEnd.y := baseList[i].locY;
      tempPath := getPath(pStart, pEnd, true, false);
      if (length(tempPath) > 0) and (tempPath[1] <> '!') then begin
        Result := pEnd;
        path := tempPath;
        highTime := baseList[i].lastChecked;
        found := true;
      end;
    end;
  end;
  // WriteLn('Found new Home Base? ' + boolToStr(found));
end;

procedure TMap.findDeliveryDist(tempPacks: TPackages);
var
  dest: Point;
  path: String;
begin
  dest := tempPacks.getNextDest();
  while (dest.x >= 0) do begin
    path := getPath(myRobot.location, dest, true, false);
    if (path[1] = '!') then
      tempPacks.setDistance(dest, -1)
    else
      tempPacks.setDistance(dest, length(path));
    dest := tempPacks.getNextDest();
  end;
end;

{----------------------- Nice Display Output --------------------- }
function TMap.getDisplayLine1(): String;
begin
  Result := 'Location = ' +
    IntToStr(myRobot.location.x) + ',' + IntToStr(myRobot.location.y);
  padString(Result, 39);
  Result := Result + 'Status: ' + myRobot.getStatus();
end;

function TMap.getDisplayLine2(): String;
begin
  Result := 'Last Command: ' + myRobot.lastCommand;
  padString(Result, 39);
  Result := Result + 'Action: ' + myRobot.lastAction;
end;

function TMap.getDisplayLine3(): String;
begin
  Result := 'Money: ' + IntToStr(myRobot.cash);
  padString(Result, 39);
  Result := Result + 'On Path: ' + boolToStr(myRobot.onPath);
end;

function TMap.getDisplayLine4(): String;
begin
  Result := 'Destination: ' + IntToStr(myRobot.dest.x) + ',' +
        IntToStr(myRobot.dest.y);
  padString(Result, 39);
  Result := Result + 'Score: ' + IntToStr(myRobot.score);
end;

function TMap.getDisplayLine5(): String;
var x,y: Integer;
begin
  x := myRobot.location.x;
  y := myRobot.location.y;
  Result := 'Packages: ' + myRobot.packages;
  padString(Result, 39);
  if (mapData[y][x].myPacks <> nil) then
    Result := Result + '# at Loc: ' + IntToStr(mapData[y][x].myPacks.getCount())
  else
    Result := Result + '# at Loc: 0';
end;

function TMap.getDisplayLine6(): String;
begin
  Result := 'Path: (' + IntToStr(Length(myRobot.path)) + ') ' + myRobot.path;
  if (length(Result) > 79) then
    setLength(Result, 79); 
end;

function TMap.getDisplayLine7(): String;
var path: String;
begin
  Result := 'Alone (1/12/Die/Kill): ' + BoolToStr(myRobot.alone) + ' (' +
        BoolToStr(myRobot.pushAlone) + '/' +
        BoolToStr(myRobot.veryAlone) + '/' +
        BoolToStr(aiCouldDie()) + '/' +
        BoolToStr(aiPushToOblivion(path)) + ')';
  padString(Result, 39);
  Result := Result + 'Oponents: ' + IntToStr(myRobot.oponents) + '/' +
        IntToStr(length(robotList));
end;

function TMap.getDisplayLine8(): String;
begin
  Result := 'Delivery Cost (Last): ' + IntToStr(myRobot.getAvgDeliverCost) +
        ' (' + IntToStr(myRobot.tripCost) + ')';
  padString(Result, 39);
  Result := Result + 'Push Count: ' + IntToStr(myRobot.pushCount);
end;

function TMap.getDisplayLine9(): String;
begin
  if (myRobot.gathering) then begin
    Result := 'Gather Dest (' + IntToStr(myRobot.gatherDest.x) + ',' +
        IntToStr(myRobot.gatherDest.y) + ')';
    padString(Result, 39);
    Result := Result + 'Gather Point: ' +
      boolToStr(mapData[myRobot.location.y][myRobot.location.x].gatherPoint);
  end else
    Result := '';
end;

procedure TMap.padString(var line: String; len: Integer);
var
  curLen: Integer;
  i: Integer;
begin
  curLen := Length(line);
  if (curLen > len) then
    line := copy(line, 1, len)
  else begin
    for i := curLen to len do
      line := line + ' ';
  end;
end;

procedure TMap.findShores();
var i, j: Integer;
begin
  for i := 0 to height do begin
    for j := 0 to width do begin
      if (mapData[i][j].flag = FLAG_PLAIN) and isShore(j, i) then
        mapData[i][j].flag := FLAG_SHORE;
    end;
  end;
end;

function TMap.isShore(x, y: Integer): Boolean;
begin
  if (x-1 >= 0) and (mapData[y][x-1].flag = FLAG_WATER) then
    Result := true
  else if (x+1 <= width) and (mapData[y][x+1].flag = FLAG_WATER) then
    Result := true
  else if (y-1 >= 0) and (mapData[y-1][x].flag = FLAG_WATER) then
    Result := true
  else if (y+1 <= height) and (mapData[y+1][x].flag = FLAG_WATER) then
    Result := true
  else
    Result := false;
end;

procedure TMap.aiFindPackage();
var
  dest: Point;          // Destination Point
  path: String;         // Walking Path
  pList: String;        // Package List
  pwList: String;       // Package Weight List
  pWeight: Integer;     // Total Package Weight
  tempPacks: TPackages;
begin
  if (not myRobot.haveDest) then begin
    // Get the point of the closest known package
    path := '';
    dest := findClosestPackage(myRobot.location, path);
    if (dest.x >= 0) and (dest.y >= 0) then begin
      myRobot.setDest(dest, path);
    end;
  end
  else begin
    // Check to see if we're at out dest, if so it's time to pick
    // up packages.
    if (myRobot.atDest) then begin
      // writeLn('Time to pick up Packages...');
      tempPacks := mapData[myRobot.location.y][myRobot.location.x].myPacks;
      if (tempPacks <> nil) then begin
        pList := tempPacks.getBestPackageList(myRobot.capacity, dest, pwList, pWeight,
                (mapData[myRobot.location.y][myRobot.location.x].flag = FLAG_HOME));
        pkSet := tempPacks.debug;
        if (length(pList) > 0) then begin
          path := getPath(myRobot.location, dest, true, false);
          myRobot.setDelivery(dest, path, pList, pwList);

          // If we're very alone (i.e. no bot within 12 spaces)
          // hide some packages if the path is long enough
          if (length(path) > ((width+height) shr 2)) and
             (not mapData[myRobot.location.y][myRobot.location.x].gatherPoint) and
             (myRobot.veryAlone) and (myRobot.shouldGather) then begin
            // If we're already gathering make sure it's within 6
            if (not myRobot.gathering) or (length(path) <= (hideDistance+3)) then begin
              myRobot.state := ROBOT_HIDE_PUT;
              RobotAI();
            end
            else begin
              myRobot.setStopGathering();
              myRobot.state := ROBOT_FINDPACKAGE;
              RobotAI();
            end;
          end;
        end
        else begin
          mapData[myRobot.location.y][myRobot.location.x].packages := false;
          myRobot.haveDest := false;
          RobotAI();
        end;
      end
      // Uh oh, no packages where they should be, time to find another place
      else begin
        // We already waited and still nothing, find them elsewhere
        if (myRobot.waitOnPackage) then begin
          mapData[myRobot.location.y][myRobot.location.x].packages := false;
          myRobot.haveDest := false;
          RobotAI();
        end
        // Wait one turn for them to come.
        else begin
          myRobot.move := 'Drop';
          myRobot.moveData := '-1';
        end;
      end;
    end
    // If the last move to our dest was on path, just keep on trucking
    else if (myRobot.checkOnPath) then begin
      myRobot.onPathMove();
    end
    // If not, then we should once again find our closest package.
    else begin
      path := '';
      dest := findClosestPackage(myRobot.location, path);
      if (dest.x >= 0) and (dest.y >= 0) then begin
        myRobot.setDest(dest, path);
      end;
    end;
  end;
end;

procedure TMap.aiDeliverPackage();
var
  path: String;         // Walking Path
begin
  // Make sure we have something to deliver if not, go find something
  if (myRobot.getWeight() <= 0) then begin
    myRobot.state := ROBOT_FINDPACKAGE;
    RobotAI();
  end
  // Make sure we have a path, if we don't find one
  else if (length(myRobot.path) > 0) and (myRobot.path[1] = '!') then begin
    path := getPath(myRobot.location, myRobot.dest, true, false);
    myRobot.setDest(myRobot.dest, path);
  end
  // Make sure we have a destination
  else if (not myRobot.haveDest) then
    myRobot.setDest(myRobot.dest, myRobot.path)
  // If we're at the destination deliver the packages
  else if (myRobot.atDest) then begin
    // If we're in gather mode be sure to set the flag
    if (myRobot.state = ROBOT_HIDE_PUT) then
      mapData[myRobot.location.y][myRobot.location.x].gatherPoint := true;
    myRobot.dropDelivery();
  end
  // If we're still on the path just keep on trucking
  else if (myRobot.checkOnPath) then
    myRobot.onPathMove()
  // If we've moved off the path, find a new one
  else begin
    path := getPath(myRobot.location, myRobot.dest, true, false);
    myRobot.setDest(myRobot.dest, path);
  end;
end;

procedure TMap.aiGatherPut();
var
  dest: Point;
  oldPath: String;
  path, pList, pwList: String;
  i, j: Integer;
  pWeight: Integer;
  tempPacks: TPackages;
begin
  dest.x := -1;
  dest.y := -1;
  path := 'find';
  setLength(path, hideDistance+1);

  // If we need a gather point, find it!
  if (not myRobot.haveGatherDest) then begin
    // We gathering we want to find the _worst_ path to lessen the
    // chance the someone will find our pile.

    // Reset our temp walls
    for i := 0 to height do begin
      for j := 0 to width do begin
        mapData[i][j].tempBlock := mapData[i][j].gatherPoint;
      end;
    end;

    while (length(path) > hideDistance) and (path[1] <> '!') do begin
      oldPath := path;
      path := getPath(myRobot.location, myRobot.dest, false, true);
      dest := getWalk(myRobot.location, Copy(path, 1, hideDistance));
      mapData[dest.y][dest.x].tempBlock := true;
    end;
    if (length(oldPath) > 0) and (oldPath[1] <> '!') then begin
      dest := getWalk(myRobot.location, Copy(oldPath, 1, hideDistance));
      path := getPath(myRobot.location, dest, true, false);
      myRobot.setDeliveryHide(dest, path);
    end
    else begin
      pList := tempPacks.getBestPackageList(myRobot.capacity, dest, pwList, pWeight,
        (mapData[myRobot.location.y][myRobot.location.x].flag = FLAG_HOME));
      pkSet := tempPacks.debug;
      if (length(pList) > 0) then begin
        path := getPath(myRobot.location, dest, true, false);
        myRobot.setDelivery(dest, path, pList, pwList);
      end;
    end;
  end
  else if (not myRobot.gathering) then begin
    path := getPath(myRobot.location, myRobot.gatherDest, true, false);
    myRobot.setDeliveryHide(myRobot.gatherDest, path);
  end
  else begin
    // After that it's just like normal delivery
    aiDeliverPackage();
  end;
end;

procedure TMap.aiOblivion();
var
  Path: String;
  dest: Point;
begin
  Path := Copy(myRobot.path, 1, 1);
  dest := GetWalk(myRobot.location, Path);
  // If we're done with the push, stop it!
  if ((myRobot.oblivionPush) or (myRobot.state = ROBOT_OBLIVION)) and
     ((length(myRobot.path) <= 0) or
      (mapData[dest.y][dest.x].flag = FLAG_WATER)) then begin
    myRobot.setOblivionOff();
    RobotAI();
  end ;

  // If we're not alone check to see if we can push someone to their death.
  // Mean, aren't I?
  if (not myRobot.pushAlone) and (aiPushToOblivion(path)) then begin
    dest := getWalk(myRobot.location, Path);
    myRobot.setOblivionOn(dest, path);
    if (length(path) = 1) then
      myRobot.bid := myRobot.oblivionKill + myRobot.oblivionStep
    else
      myRobot.bid := myRobot.oblivionStep;
  end;
end;

procedure TMap.aiBidLevel();
begin
  // If we're alone, don't worry about the bid
  if (myRobot.alone) then begin
    myRobot.bid := 1;
    push2Count := 0;
  end
  // If we can push/be pushed we need to do some extra caculating.
  else if (not myRobot.pushAlone) then begin
    // If we have packages bid higher.  If we've been pushed, bid
    // even higher.

    // If we can be pushed into water this turn, bid high!
    if aiCouldDie() then begin
      myRobot.bid := myRobot.getAvgDeliverCost() + (myRobot.cash shr 4) + 20;
      myRobot.move := 'Move';
      myRobot.moveData := aiGetSafeDir();
    end
    // If we're getting pushed around when it doesn't matter bid low
    else if aiPushBlocked() then
      myRobot.bid := 1
    else if (length(myRobot.packages) > 0) then
      myRobot.bid := 2 + myRobot.pushCount
    // If we're being try something else
    else if (myRobot.pushCount > 2) then begin
      myRobot.setReset();
      RobotAI();
      push2Count := 0;
    end
    // If all else fails, bid a little higher.
    else begin
      myRobot.bid := 2;
      inc(push2Count);
      // If we've been pushing too long, find another way
      if (push2Count > (((width+height) shr 5) + 1)) then begin
        myRobot.setReset();
        RobotAI();
        push2Count := 0;
      end;
    end
  end
  // If we have another robot within 2 turns bid low (-1) so they move
  // first (hopefully), that way we move after and have a pushing chance
  else if (myRobot.pushAlone) and (not myRobot.alone) then begin
    myRobot.bid := -1;
  end
  else begin
    myRobot.bid := 1;
  end;
end;

function TMap.aiCouldDie(): Boolean;
var
  locX, locY: Integer;
  N,E,S,W: String;
begin
  locX := myRobot.location.x;
  locY := myRobot.location.y;
  Result := false;
  N := getMapChar(locX, locY-1, true);
  S := getMapChar(locX, locY+1, true);
  E := getMapChar(locX+1, locY, true);
  W := getMapChar(locX-1, locY, true);
  if (N = 'w') and ((S = 'R') or (S = '%')) then
    Result := true
  else if (S = 'w') and ((N = 'R') or (N = '%')) then
    Result := true
  else if (E = 'w') and ((W = 'R') or (W = '%')) then
    Result := true
  else if (W = 'w') and ((E = 'R') or (E = '%')) then
    Result := true;
end;

function TMap.aiPushBlocked(): Boolean;
var
  locX, locY: Integer;
  N,E,S,W: String;
begin
  locX := myRobot.location.x;
  locY := myRobot.location.y;
  Result := false;
  N := getMapChar(locX, locY-1, true);
  S := getMapChar(locX, locY+1, true);
  E := getMapChar(locX+1, locY, true);
  W := getMapChar(locX-1, locY, true);
  if (N = '#') and ((S = 'R') or (S = '%')) then
    Result := true
  else if (S = '#') and ((N = 'R') or (N = '%')) then
    Result := true
  else if (E = '#') and ((W = 'R') or (W = '%')) then
    Result := true
  else if (W = '#') and ((E = 'R') or (E = '%')) then
    Result := true;
end;

function TMap.aiPushToOblivion(var Path: String): Boolean;
var
  locX, locY: Integer;
  N,E,S,W: String;
  Read: String;
  Dir: String;
  i, count: Integer;
begin
  locX := myRobot.location.x;
  locY := myRobot.location.y;
  Result := false;
  Count := 0;
  S := getMapChar(locX, locY-1, true);
  N := getMapChar(locX, locY+1, true);
  E := getMapChar(locX+1, locY, true);
  W := getMapChar(locX-1, locY, true);
  if (S = '%') or (S = 'R') then begin
    i := locY-2;    count := 0;
    Read := getMapChar(locX, i, true);
    while (Read = ' ') do begin
      dec(i);
      inc(count);
      Read := getMapChar(locX, i, true);
    end;
    if (Read = 'w') then begin
      Result := true;
      Dir := 'S';
    end;
  end;
  if (not Result) and ((N = '%') or (N = 'R')) then begin
    i := locY+2;    count := 0;
    Read := getMapChar(locX, i, true);
    while (Read = ' ') do begin
      inc(i);
      inc(count);
      Read := getMapChar(locX, i, true);
    end;
    if (Read = 'w') then begin
      Result := true;
      Dir := 'N';
    end;
  end;
  if (not Result) and ((E = '%') or (E = 'R')) then begin
    i := locX+2;    count := 0;
    Read := getMapChar(i, LocY, true);
    while (Read = ' ') do begin
      inc(i);
      inc(count);
      Read := getMapChar(i, LocY, true);
    end;
    if (Read = 'w') then begin
      Result := true;
      Dir := 'E';
    end;
  end;
  if (not Result) and ((W = '%') or (E = 'R')) then begin
    i := locX-2;    count := 0;
    Read := getMapChar(i, LocY, true);
    while (Read = ' ') do begin
      dec(i);
      inc(count);
      Read := getMapChar(i, LocY, true);
    end;
    if (Read = 'w') then begin
      Result := true;
      Dir := 'W';
    end;
  end;

  If (Result) then begin
    Path := '';
    for i := 0 to count do
      Path := Path + Dir;
  end;
end;

function TMap.aiGetSafeDir(): String;
var
  locX, locY: Integer;
  N,E,S,W: String;
begin
  Result := '';
  locX := myRobot.location.x;
  locY := myRobot.location.y;
  S := getMapChar(locX, locY-1, true);
  N := getMapChar(locX, locY+1, true);
  E := getMapChar(locX+1, locY, true);
  W := getMapChar(locX-1, locY, true);
  // Nothing would be best
  if (S = ' ') then
    Result := 'S'
  else if (N = ' ') then
    Result := 'N'
  else if (W = ' ') then
    Result := 'W'
  else if (E = ' ') then
    Result := 'E'
  // Robot with Package would be second best
  else if (S = '%') then
    Result := 'S'
  else if (N = '%') then
    Result := 'N'
  else if (W = '%') then
    Result := 'W'
  else if (E = '%') then
    Result := 'E'
  // Robot is okay too
  else if (S = 'R') then
    Result := 'S'
  else if (N = 'R') then
    Result := 'N'
  else if (W = 'R') then
    Result := 'W'
  else if (E = 'R') then
    Result := 'E'
  // We should never get here, but just in case, default to North
  else
    Result := 'N';
end;


procedure TMap.addRobot(id: Integer);
var loc: Integer;
begin
  if (length(robotList) = 0) then begin
    loc := 0;
    setLength(robotList, loc + 1);
    robotList[loc].Id := id;
    robotList[loc].LocX := -1;
    robotList[loc].LocY := -1;
    robotList[loc].isAlive := true;
    robotList[loc].packageCount := 0;
  end else begin
    loc := High(robotList) + 1;
    setLength(robotList, loc + 1);
    robotList[loc].Id := id;
    robotList[loc].LocX := -1;
    robotList[loc].LocY := -1;
    robotList[loc].isAlive := true;
    robotList[loc].packageCount := 0;
  end;
end;

function TMap.haveRobot(id: Integer): Boolean;
var i: Integer;
begin
  Result := false;
  if (length(robotList) > 0) then begin
    for i := low(robotList) to high(robotList) do begin
      if (robotList[i].Id = id) then
        Result := true;
    end;
  end;
end;

procedure TMap.robotSetX(id, x: Integer);
var i: Integer;
begin
  if (length(robotList) > 0) then begin
    for i := low(robotList) to high(robotList) do begin
      if (robotList[i].Id = id) then begin
        robotList[i].LocX := x - 1;
        if (robotList[i].LocX >= 0) and (robotList[i].LocY >= 0) then
          mapData[robotList[i].LocY][robotList[i].LocX].robot := id;
      end;
    end;
  end;
end;

procedure TMap.robotSetY(id, y: Integer);
var i: Integer;
begin
  if (length(robotList) > 0) then begin
    for i := low(robotList) to high(robotList) do begin
      if (robotList[i].Id = id) then begin
        robotList[i].LocY := y - 1;
        if (robotList[i].LocX >= 0) and (robotList[i].LocY >= 0) then
          mapData[robotList[i].LocY][robotList[i].LocX].robot := id;
      end;
    end;
  end;
end;

procedure TMap.robotMove(id: Integer; dir: Char);
var
 oldX, oldY: Integer;
 i: Integer;
begin
  if (length(robotList) > 0) then begin
    for i := low(robotList) to high(robotList) do begin
      if (robotList[i].Id = id) then begin
        oldX := robotList[i].LocX;
        oldY := robotList[i].LocY;
        case (dir) of
          STMOVE_NORTH: dec(robotList[i].locY);
          STMOVE_SOUTH: inc(robotList[i].locY);
          STMOVE_WEST:  dec(robotList[i].locX);
          STMOVE_EAST:  inc(robotList[i].locX);
        end;
        if (oldX >= 0) and (oldY >= 0) and
           (oldX <= width) and (oldY <= height) then begin
          mapData[oldY][oldX].robot := -1;
          mapData[oldY][oldX].robotPacks := -1;
        end;
        if (robotList[i].locX >= 0) and
           (robotList[i].LocX <= width) and
           (robotList[i].LocY >= 0) and
           (robotList[i].LocY <= height) then begin
          mapData[robotList[i].locY][robotList[i].LocX].robot := id;
          mapData[robotList[i].locY][robotList[i].LocX].robotPacks := robotList[i].packageCount;
        end;
      end;
    end;
  end;
end;

procedure TMap.robotPick(id, pId: Integer);
var
  i: Integer;
  tempPack: TPackages;
begin
  // It's My Robot doing the picking
  if (id = -1) then begin
    tempPack := mapData[myRobot.location.y][myRobot.location.x].myPacks;
    if (tempPack <> nil) then
      tempPack.picked(pId);
  end
  else if (length(robotList) > 0) then begin
    for i := low(robotList) to high(robotList) do begin
      if (robotList[i].Id = id) then begin
        inc(robotList[i].packageCount);
        tempPack := mapData[robotList[i].locY][robotList[i].locX].myPacks;
        // If this is a home base and we think it's empty, reset it for checking
        if (mapData[robotList[i].LocY][robotList[i].LocX].flag = FLAG_HOME) and
           (mapData[robotList[i].LocY][robotList[i].LocX].packages = false) then
          mapData[robotList[i].locY][robotList[i].LocX].packages := true;

        // fix the package info so we don't try and pick up the packages
        if (tempPack <> nil) then
          tempPack.picked(pId);
      end;
    end;
  end;
end;

procedure TMap.robotDrop(id, pId: Integer);
var
  i: Integer;
begin
  if (length(robotList) > 0) then begin
    for i := low(robotList) to high(robotList) do begin
      if (robotList[i].Id = id) then begin
        dec(robotList[i].packageCount);
        if (robotList[i].packageCount < 0) then
          robotList[i].packageCount := 0;
      end;
    end;
  end;
end;

// Robot sent a packet, it's alive!
procedure TMap.robotAlive(id: Integer);
var i: Integer;
begin
  if (length(robotList) > 0) then begin
    for i := low(robotList) to high(robotList) do begin
      if (robotList[i].Id = id) then begin
        robotList[i].isAlive := true;
        if (robotList[i].LocX >= 0) and (robotList[i].LocY >= 0) then
          mapData[robotList[i].LocY][robotList[i].LocX].robot := id;
      end;
    end;
  end;
end;

// Kill all the robots at the start of every round, if they
// send info they are still alive.
procedure TMap.robotKillAll();
var i: Integer;
begin
  if (length(robotList) > 0) then begin
    for i := low(robotList) to high(robotList) do begin
      robotList[i].isAlive := false;
      if (robotList[i].LocX >= 0) and (robotList[i].LocY >= 0) then
        mapData[robotList[i].LocY][robotList[i].LocX].robot := -1;
    end;
  end;
end;

function TMap.robotCount(): Integer;
var i: Integer;
begin
  Result := 0;
  if (length(robotList) > 0) then begin
    for i := low(robotList) to high(robotList) do begin
      if (robotList[i].id > 0) and (robotList[i].isAlive) then
        inc(Result);
    end;
  end;
end;

function TMap.getMapChar(x, y: Integer; simple: Boolean): String;
var
  map: MapObject;
begin
  if (x < 0) or (x > width) or (y < 0) or (y > height) then begin
    Result := '#';
  end
  else begin
    map := mapData[y][x];
    if (map.flag = FLAG_WATER) then
      Result := 'w'
    else if (map.flag = FLAG_WALL) then
      Result := '#'
    else if (not simple) and
            (x = myRobot.location.x) and
            (y = myRobot.location.y) then
      Result := 'M'
    else if (map.robot > 0) then begin
      if (simple and (mapData[y][x].robotPacks > 0)) then
        Result := '%'
      else if (simple) then
        Result := 'R'
      else if (mapData[y][x].robotPacks > 0) and
         (mapData[y][x].robotPacks < 10) then
        Result := IntToStr(mapData[y][x].robotPacks)
      else if (mapData[y][x].robotPacks > 9) then
        Result := '%'
      else
        Result := 'R';
    end
    else if (not simple) and (map.packages) then begin
      if (map.gatherPoint) then
        Result := 'G'
      else
        Result := '@';
    end
    else if (not simple) and (mapData[y][x].flag = FLAG_SHORE) then
      Result := '-'
    else
      Result := ' ';
  end;
end;

procedure TMap.addHomeBase(x,y: Integer);
var loc: Integer;
begin
  if (length(baseList) = 0) then begin
    loc := 0;
    setLength(baseList, loc + 1);
    baseList[loc].locX := x;
    baseList[loc].locY := y;
    baseList[loc].lastChecked := -1;
    baseList[loc].packageCount := -1;
  end else begin
    loc := High(baseList) + 1;
    setLength(baseList, loc + 1);
    baseList[loc].locX := x;
    baseList[loc].locY := y;
    baseList[loc].lastChecked := -1;
    baseList[loc].packageCount := -1;
  end;
end;

procedure TMap.setHomeBase(x, y, cTime, pCount: Integer);
var i: Integer;
begin
  for i := low(baseList) to high(baseList) do begin
    if (baseList[i].locX = x) and (baseList[i].locY = y) then begin
      baseList[i].lastChecked := cTime;
      baseList[i].packageCount := pCount;
    end;
  end;
end;

function TMap.BoolToStr(b: Boolean): String;
begin
  if (b) then
    Result := 'T'
  else
    Result := 'F';
end;

end.
